--------------------------=( Script Information )=--------------------------
-- Name :         AutoVP
-- Version :      v1.02
-- Last Changed : 2010/Apr/18
-- Built for :    3ds max 5+
--
-- --------------------------------=( Summary )=-------------------------------
-- AutoVP automatically changes the active viewport to the viewport the mouse
-- cursor is hovering over.
--
-- ---------------------------=( Copyright Notice )=---------------------------
-- by Richard Annema <rannema@splutterfish.com>
--
-- ---------------------------=( Acknowledgements )=---------------------------
-- For inspiration, the "Autoselecting viewport?" Forum Thread at CGTalk
-- started by user "Gen-an".
-- http://forums.cgsociety.org/showthread.php?f=98&t=680404
-- - URL retrieved on 2008/Oct/17.
--
-- -----------------------------=( Installation )=-----------------------------
-- Simply run this script.  AutoVP will then be available from the Customize
-- User Interface dialog in the SplutterFish category.
--
-- For permanent accessibility, place this file in the Startup scripts folder.
--
-- -----------------------------=( Documentation )=----------------------------
-- UI:
--   Place the macroscript on a toolbar (button, recommended) menu entry or
--   keyboard shortcut.
--
--   Press the button / keyboard combination / select the menu entry to enable
--   AutoVP.
--   Press again / deselect to disable AutoVP.
--
--   In versions prior to 3ds Max 9, a small dialog with an enable/disable
--   button will appear.  When this dialog is closed, AutoVP will be disabled.
--
--   In versions after 3ds Max 9, the AutoVP function is handled in the
--   background via a .NET timer.  If you prefer the dialog mode, hold the
--   Shift key when selecting AutoVP, or set the 'autoVP_useDotNetTimer'
--   setting to 'false' (without quotes) in the PREFERENCES section below.
--
--   AutoVP works largely based on a timer.  The default interval at which it
--   checks which viewport the mouse cursor is hovering over is 100
--   milliseconds. To change this interval, change the value of the
--   'autoVP_interval' setting in the PREFERENCES section below, and re-
--   -evaluate this script.
--
-- ----------------------------------------------------------------------------
-- Scripting:
--   <bool>autoVP.start [method:(#dotnet|#rollout) [rolloutCoords:<point2>]
--     Starts autoVP
--   Parameters:
--     method
--       The timer method that should be used.
--       #dotnet - uses a .NET system timer
--       #rollout - uses a maxscript rollout timer (creates rollout)
--       Default: #dotnet on 3ds Max 9+, #rollout on earlier releases
--     rolloutCoords
--       A point2 value that determines the coordinates where the rollout, if
--       method:#rollout was specified, should open.
--       You can specify off-screen coordinates to hide the rollout, and [-1,-1]
--       to open the rollout in the center of the screen.
--       Default: Center of screen on first run, last open position afterwards.
--   Returns:
--     true
--       autoVP started
--     false
--       autoVP was already running
--
--   <OK>autoVP.stop()
--     Stops AutoVP
--   Returns:
--     OK
--
--   <undefined|<+int:index>autoVP.getViewportUnderMouse()
--     Gets the index of the viewport the mouse is hovering over, if any.
--     Call this function when you have a script that needs this information.
--   Returns:
--     undefined
--       The mouse is not hovering over a viewport.
--     index
--       The index of the viewport the mouse is hovering over.  This
--       corresponds with the viewport indices as used in MaxScript.
--
-- ---------------------------=( Revision History )=---------------------------
-- v1.02
--   - Added a callback function to determine whether AutoVP should be
--     temporarily disabled.
--     The callback function pointer is stored in autoVP_disableCB.
--   - Added two example callback functions:
--     - autoVP_disableCB_LMB (script default)
--       Disables AutoVP if the left mouse button is pressed.  This prevents
--       odd behavion drag-moving/rotating/scaling objects where the mouse
--       cursor exits the originating viewport's bounds.
--     - autoVP_disableCB_Shift
--       Disables AutoVP when the Shift key is pressed.  Can be used for
--       similar purposes.
--   - To create your own callback, just create a function that returns a
--     booleanClass value.  true = AutoVP ON. false = AutoVP OFF.
--     Make sure that if you define a custom callback, that it is defined
--     before AutoVP is loaded.  This typically means adding your custom
--     callback function to the "CALLBACK FN" section of the script, or adding
--     'global functionName' just above the autoVP_disableCB variable.
--
--   - AutoVP can now be temporarily disabled by holding the Shift key.
--     Through this, the undesirable behavior when drag-moving / drag-rotating
--     objects where the mouse cursor leaves the current viewport can be
--     prevented.  Please make sure to press the Shift key -after- beginning
--     such an operation to prevent the move/rotate turning into a cloning
--     operation.
--    
--
-- v1.01
--  - Made .NET vs dialog toggle an option by holding the Shift key
--  - Separated methods from macrosript so they can be accessed externally.
--
-- v1.0 - first full release version
--
-- -----------------------------=( Known Issues )=-----------------------------
-- As it appears to be impossible to detect whether the main 3ds Max 'window'
-- has focus or not, AutoVP will continue to work even if working in another
-- application, another dialog in 3ds Max, or when working with the viewport
-- settings right-click menu.  This is typically harmless - however, care
-- should be taken in the latter situation that the viewport settings are not
-- accidentally applied to a different viewport that was auto-selected.
--
-- ----------------------------------------------------------------------------

-- -----------------------------=( CALLBACK FN )=------------------------------
-- This script must be re-evaluated after changing a callback function!
-- ----------------------------------------------------------------------------
fn autoVP_disableCB_LMB = ( not mouse.buttonStates[1] )
fn autoVP_disableCB_Shift = ( not keyboard.ShiftPressed )
-- -----------------------------=( /CALLBACK FN )=-----------------------------

-- ------------------------------=( PREFERENCES )=-----------------------------
-- This script must be re-evaluated after changing a preference!
-- ----------------------------------------------------------------------------
	global autoVP_method = #dotnet -- toggles button behavior - this is the method used when shift is not pressed
	global autoVP_interval = 100 -- millisecond interval at which to poll the mouse position
	global autoVP_disableCB = autoVP_disableCB_LMB -- 'temporarily disable' callback function to use
-- ------------------------------=( /PREFERENCES )=----------------------------

-- globals
global autoVP, roll_autoVP, autoVP_dotNetInvoke

-- --------------------------------=( SCRIPT )=--------------------------------
-- Main struct with variables and functions
Struct AutoVP_s (
	viewportBoxes,
	viewportLayout,
	viewportCount,
	dotNetTimer = undefined,
	
	-- initialization function... get the positions/dimensions for each viewport
	fn init = (
		viewportBoxes = #()
		local currentVp = viewport.activeViewPort
		for i = 1 to viewport.numViews do (
			local index = (mod (i + currentVp) viewport.numViews) + 1
			viewport.activeViewPort = index
			local vpTopLeft = mouse.screenPos - mouse.pos
			local vpSize = getViewSize()
			viewportBoxes[index] = (Box2 vpTopLeft.x vpTopLeft.y vpSize.x vpSize.y)
		)
		if (currentVp != 0) then ( viewport.activeViewPort = currentVp )
		viewportLayout = viewport.getLayout()
		viewportCount = viewport.numViews
	),
		
	fn getViewportUnderMouse = (
		local skip = false
		-- better make sure that layout changes don't ruin things here
		if (viewport.getLayout() != autoVP.viewportLayout) then (
			autoVP.init()
		)
		else if (viewport.numViews != autoVP.viewportCount) then (
			autoVP.init()
		)
		else (
			-- ditto for viewport resizings
			local currentVp = viewport.activeViewPort
			if (currentVp != 0) then (
				local currentVpAssumedBox = autoVP.viewportBoxes[currentVp]
				if (getViewSize() != [currentVpAssumedBox.w,currentVpAssumedBox.h]) then (
					autoVP.init()
				)
			)
		)

		local result = undefined
		for i = 1 to viewport.numViews do (
			viewportBox = autoVP.viewportBoxes[i]
			if (contains viewportBox mouse.screenPos) then (
				result = i
				exit
			)
		)
		result
	),

	fn setMouseActiveViewport autoVP_disableCB:autoVP_disableCB = (
		if ((classOf autoVP_disableCB == MaxScriptFunction) AND (autoVP_disableCB())) do (
			local hoverView = getViewportUnderMouse()
			if ((hoverView != undefined) AND (hoverView != viewport.activeViewport)) then (
				viewport.activeViewPort = hoverView
			)
		)
	),

	active = false,

	fn start method:#dotnet rolloutCoords: = (
		if (autoVP.active) then ( false )
		autoVP.active = true
		if (method == #dotnet) then (
			if (autoVP.dotNetTimer == undefined) then (
				autoVP.dotNetTimer = dotNetObject "System.Windows.Forms.Timer"
				dotnet.addEventHandler autoVP.dotNetTimer "tick" autoVP_dotNetInvoke
				autoVP.dotNetTimer.interval = autoVP_interval
			)
			autoVP.dotNetTimer.start()
		)
		else (
			if (classOf roll_autoVP == rolloutClass) then ( destroyDialog roll_autoVP )
			if (rolloutCoords != unsupplied) then ( roll_autoVP.uiPos = rolloutCoords )
			else if (roll_autoVP.uiPos == undefined) then ( roll_autoVP.uiPos = [-1,-1] )
			createDialog roll_autoVP roll_autoVP.width roll_autoVP.height roll_autoVP.uiPos.x roll_autoVP.uiPos.y style:#(#style_toolwindow,#style_sysmenu)
		)
		true
	),
	fn stop = (
		if (autoVP.dotNetTimer != undefined) then (
			if (autoVP.dotNetTimer.enabled) then ( autoVP.dotNetTimer.stop() )
		)
		if (roll_autoVP.open) then ( destroyDialog roll_autoVP )
		autoVP.active = false
		OK
	)
)
-- in case of re-evaluation, kill any existing autoVP .NET timer
if ((autoVP != undefined) AND (autoVP.dotNetTimer != undefined)) then (
	autoVP.dotNetTimer.stop()
	autoVP.dotNetTimer = undefined
)

-- .NET events can't invoke functions that are within Structs properly ... so here's an intermediating function >_<
fn autoVP_dotNetInvoke = (
	try (
		autoVP.setMouseActiveViewport()
	)
	catch (
		autoVP.dotNetTimer.stop()
		local err = getCurrentException()
		messageBox ("An error occurred and AutoVP has been halted.  The error was:\r\n" + err)
	)
)

autoVP = autoVP_s()

-- UI
if (classOf roll_autoVP == RolloutClass) then ( destroyDialog roll_autoVP )
rollout roll_autoVP "autoVP" width:64 height:32 (
	local uiPos
	
	timer ticktock interval:100 active:true
	checkbutton ckb_active "active" checked:true pos:[4,4] width:56 height:24 highlightColor:(color 192 255 192) tooltip:"Toggle AutoVP"
	
	on ticktock tick do ( autoVP.setMouseActiveViewport() )
	
	on ckb_active changed state do (
		ticktock.active = state
		autoVP.active = state
		updateToolbarButtons()
	)
	
	on roll_autoVP open do (
		uiPos = getDialogPos roll_autovp
		ticktock.interval = autoVP_interval
		if (autoVP.dotNetTimer != undefined) then ( autoVP.dotNetTimer.enabled = false )
		autoVP.init()
		autoVP.active = true
	)
	on roll_autoVP moved pos do ( uiPos = pos )
	on roll_autoVP close do (
		autoVP.active = false
		updateToolbarButtons()
	)
)

macroscript AutoVP_mcr category:"SplutterFish" buttontext:"AutoVP" tooltip:"AutoVP - MouseOver Viewport Switcher (v1.00)" (
	on execute do (
		if (autoVP == undefined) do (
			messageBox "Please re-install autoVP" title:"autoVP is not installed"
			return false
		)
		if (autoVP_method == undefined) then ( global autoVP_method = #dotnet )
		local method = (
			if ((maxVersion())[1] < 9000) then ( #rollout ) -- no .NET in 3ds Max releases below 9
			else if (keyboard.shiftPressed) then ( -- flip behavior
				if (autoVP_method == #dotnet) then ( #rollout )
				else ( #dotnet )
			)
			else ( autoVP_method )
		)
		if (autoVP.active) then ( autoVP.stop() )
		else ( autoVP.start method:method )
		updateToolbarButtons()
	)

	on isChecked do (
		if (autoVP != undefined) then ( autoVP.active )
		else ( false )
	)
)
-- --------------------------------=( /SCRIPT )=-------------------------------